/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef MESSAGE_TRANSCEIVER_THREAD_H
#define MESSAGE_TRANSCEIVER_THREAD_H

#include <app/Message.h>
#include <app/Messenger.h>
#include <kernel/OS.h>
#include <support/Locker.h>
#include "iogateway/AbstractMessageIOGateway.h"
#include "besupport/ThreadSafeObjectPool.h"

#define PORTABLE_MESSAGES_RECEIVED 'PmRR'

//:Type representing any function that will return a newly allocated AbstractMessageIOGateway object that uses the given PortableDataIO object.
typedef AbstractMessageIOGateway* (*GatewayFactoryFunc)(PortableDataIO * io);

//:This is a utility class to help interface multithreaded BeOS
//:code more easily to a TCP socket that is sending and receiving
//:PortableMessages.   Only compiles under BeOS!
// It helps by allowing you to send and receive
// PortableMessages without having to worry about calling select(),
// or any blocking issues.
//
// Usage:  1. Create a new MessageTransceiverThread object.
//         2. Call StartThread(), StartAcceptThread(), or StartConnectThread().
//         3. Whenever you wish to send a PortableMessage over the TCP socket,
//            simply call SendOutgoingMessage().  
//         4. Whenever a PortableMessage is received, a PORTABLE_MESSAGES_RECEIVED
//            BMessage will be sent to the BMessenger you specified in StartThread().
//            When you get this, you can respond by calling GetNextIncomingThread()
//            on this object to retrieve the PortableMessage.
//         5. When you are done, just delete this object normally.  If you wish,
//            you can call WaitForAllMessagesToBeSent() first, in order to ensure
//            that all pending outbound messages get delivered.
//         6. That's it!
//
class MessageTransceiverThread
{
public:
   // Creates a receiver thread without a reply target.  You should
   // call SetReplyTarget() before calling StartThread()...
   MessageTransceiverThread(int32 threadPri = B_NORMAL_PRIORITY);

   // (replyTarget) specified where you want the PORTABLE_MESSAGES_RECEIVED (and other messages) sent to .
   // (threadPri) is the priority the TCP thread should run at.
   MessageTransceiverThread(const BMessenger & replyTarget, int32 threadPri = B_NORMAL_PRIORITY);

   // May be called at any time
   ~MessageTransceiverThread();

   // Creates a TCP send/receive thread that connects to the given hostname and port.
   // (hostname) is the Internet hostname of the computer to connect to.
   // (port) is the port number to connect on.
   // (optConnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                      is complete.  (A copy of this message is made).
   // (optDisconnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                         fails or is broken. (A copy of this message is made).
   status_t StartConnectThread(const char * hostname, int port, 
                               const BMessage * optConnectMessage = NULL, const BMessage * optDisconnectMessage = NULL);

   // Creates a TCP send/receive thread that waits for connections on the given port.
   // If (port) is -1, then the system will choose a port to accept on.  The port number it
   // chooses can be discovered by calling GetPort().
   // (acceptMultiple), if true, will allow subsequent connections to be accepted after the first one is closed.
   //                   In no case will there be more than one connection active at a time, however.
   // (optConnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                      is complete.  (A copy of this message is made).
   // (optDisconnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                         fails or is broken. (A copy of this message is made).
   status_t StartAcceptThread(int port, bool acceptMultiple,
                              const BMessage * optConnectMessage = NULL, const BMessage * optDisconnectMessage = NULL);

   // Creates and activates the TCP send/receive background thread.
   // (gateway) is the AbstractMessageIOGateway the thread should use to send and receive data.
   //           In general, a AbstractMessageIOGateway object with a non-blocking TCPSocketDataIO
   //           that uses your socket is the way to go here.
   // (optConnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                      is complete.  (A copy of this message is made).
   // (optDisconnectMessage), if specified, will be sent via your BMessenger when the connection 
   //                         fails or is broken. (A copy of this message is made).
   status_t StartThread(AbstractMessageIOGateway * gateway, int datafd, const BMessage * optDisconnectMessage = NULL);
   
   // If there are any incoming messages pending, places a reference to the next one into (setRef) and returns B_NO_ERROR.
   // Otherwise returns B_ERROR.
   status_t GetNextIncomingMessage(PortableMessageRef & setRef);

   // Enqueues (msgRef) for immediate output over the TCP connection.
   // Returns B_NO_ERROR on success, B_ERROR if no more messages can be sent (e.g. connection is broken)
   // Note that it is NOT safe to keep your own local copies of (msgRef) around
   // after this method returns (because PortableMessageRef isn't thread safe)
   // NOTE:  If you want to use object pools with your (msgRef), make sure that the
   //        pools you use are ThreadSafeObjectPools objects, or you will suffer from
   //        nasty race conditions.  The GetThreadSafeMessagePool() and GetThreadSafeMessageRefPool()
   //        methods in this class return appropriate pool objects to use here.
   status_t SendOutgoingMessage(const PortableMessageRef & msgRef);

   // Waits until all messages that were on the queue to be sent are sent, and then returns B_NO_ERROR.
   // If for some reason it could not wait, B_ERROR is returned.
   void WaitForAllMessagesToBeSent();

   // Returns the port number that the accept thread is accepting messages on. 
   // This number is only meaningful if StartAcceptThread() was called.
   int GetPort() const;

   // Sets or resets the reply target for this thread
   void SetReplyTarget(const BMessenger & target);

   // If you wish to use a type of AbstractMessageIOGateway other than the default type,
   // you can set a callback to your own factory function here.  The factory function should
   // return a new AbstractMessageIOGateway object (or subclass thereof), or NULL on failure.
   // The returned object becomes property of this MessageTransceiverThread object.
   void SetGatewayFactoryFunc(GatewayFactoryFunc func);

   // Returns a pointer to a singleton thread-safe message pool.
   // Equivalent to ::GetMessagePool(), but for use in multithreaded BeOS environments.
   static ThreadSafeObjectPool<PortableMessage> * GetMessagePool() {return &_msgPool;}

   // Returns a pointer to a singleton thread-safe message ref pool.
   // Equivalent to ::GetMessageRefPool(), but for use in multithreaded BeOS environments.
   static ThreadSafeObjectPool<PortableMessageRef::RefCountMem> * GetMessageRefPool() {return &_msgRefPool;}

private:
   int HackBind(int s, struct sockaddr * addr, int size) const;
   AbstractMessageIOGateway * CreateIOGateway(int socketfd) const;

   status_t StartThread(thread_func func, const char * threadName);
   void NotifyThread();
   status_t SetSocketBlockingEnabled(int socket, bool blocking);
   status_t CreateNotificationMechanism();
   static int32 TCPThreadStub(void * d) {return ((MessageTransceiverThread *)d)->TCPThreadLoop();}
   static int32 ConnectThreadStub(void * d) {return ((MessageTransceiverThread *)d)->ConnectThreadLoop();}
   int32 TCPThreadLoop();
   int32 ConnectThreadLoop();

   volatile thread_id _tcpThread;
   BMessenger _replyTarget;
   BMessage _notifyMsg;
   BLocker _lock;   

   // _gateway and _datafd are linked.  If _gateway if non-NULL, _datafd must be a valid fd.
   // And if _gateway is NULL, _datafd must be < 0.
   AbstractMessageIOGateway * _gateway;
   int _datafd;    // the fd we actually transfer messages over

   int _acceptfd;  // the fd we wait for connections on (or -1 if we aren't doing that)

   int _notifyfd;  // the fd we use to wake up our thread (hacksville!)
   int _waitfd;    // the fd the thread uses to be woken up (hack hack!)

   PortableString _connectTo;  // used by StartConnectThread()
   int _port;
   int32 _threadPri;
 
   bool _isConnecting;

   BMessage _onConnect;
   bool _sendOnConnect;

   BMessage _onDisconnect;
   bool _sendOnDisconnect;

   bool _acceptMultiple;

   BLocker _messagesPendingLock;  // locked whenever outgoing messages are pending

   GatewayFactoryFunc _userGatewayFunc;

   static ThreadSafeObjectPool<PortableMessage> _msgPool;
   static ThreadSafeObjectPool<PortableMessageRef::RefCountMem> _msgRefPool;
};

#endif
