/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef ABSTRACT_MESSAGE_IO_GATEWAY
#define ABSTRACT_MESSAGE_IO_GATEWAY

#include "dataio/PortableDataIO.h"
#include "message/PortableMessage.h"
#include "queue/PortableQueue.h"
#include "refcount/PortableRefCount.h"

//:Interface representing an object that can send/receive PortableMessages via a PortableDataIO byte-stream.
class AbstractMessageIOGateway
{
public:
   // This object becomes the owner of (io) and will delete it later.
   AbstractMessageIOGateway(PortableDataIO * io);
   virtual ~AbstractMessageIOGateway();

   // Adds the given message reference to our list of outgoing messages
   // to send.  Never blocks.  Returns B_ERROR iff for some reason
   // the message can't be queued (out of memory?)
   status_t AddOutgoingMessage(const PortableMessageRef & messageRef);

   // Retrieves the next message from the incoming messages queue.
   // Returns B_ERROR iff the queue is empty (in which case setRef
   // will not be modified) 
   status_t GetNextIncomingMessage(PortableMessageRef & setRef);

   // Writes some of our outgoing messages to the wire.
   // Not guaranteed to write all outgoing messages (it will try not to block)
   // or some other catastrophic condition has occurred.
   virtual status_t DoOutput() = 0;

   // Reads some more bytes from the wire.  
   // If enough bytes have been read to assemble one or more new PortableMessages,
   // the new PortableMessages will be available after this method returns by
   // calling GetNextIncomingMessage().  
   // Tries not to block, but may (depending on implementation)
   // Returns B_NO_ERROR usually, or B_ERROR if the connection has been broken
   // or some other catastrophic condition has occurred.
   virtual status_t DoInput() = 0;

   // Returns true iff there are any PortableMessages that have been
   // added (with AddOutgoingMessage()) but not already sent across the PortableDataIO.
   // Always returns false if the connection has been closed or hosed.
   virtual bool HasBytesToOutput() const = 0;

   // Returns true iff this gateway is interested in new data arriving at its
   // socket.  Default value is true.
   bool GetInputEnabled() const;

   // Call this to enable/disable reading of new incoming data from the socket.
   void SetInputEnabled(bool enable);

   // Returns true iff there are any incoming PortableMessages ready to
   // be retrieved via GetNextIncomingMessage().
   bool HasIncomingMessagesReady() const;

   // By default, the AbstractMessageIOGateway calls Flush() on its PortableDataIO's
   // output stream whenever the last outgoing message in the outgoing message queue 
   // is sent.  Call SetFlushOnEmpty(false) to inhibit this behavior (e.g. for bandwidth
   // efficiency when low message latency is not a requirement).
   void SetFlushOnEmpty(bool flush);

   // By default, the AbstractMessageIOGateway will use the PortableMessage class's 
   // built-in message and message-ref pool objects to allocate its messages.  This
   // is okay for single-threaded programs, but can lead to race conditions in multi-threaded
   // situations.  In those cases, call SetPools() with your own separate or synchronized
   // pool objects to avoid trouble.
   // The pools may be set to NULL to disable object pooling entirely (i.e. use new/delete instead)
   void SetPools(PortableMessageRef::ItemPool * ioMessagePool, PortableMessageRef::RefPool * ioRefPool);

protected:
   // Allocates or reallocates a buffer so that when this method returns,
   // the buffer is at least (desiredSize) bytes long.
   // (bufPtr) points to a pointer to the buffer.  (May point to a NULL)
   // pointer if a buffer hasn't been allocated yet)
   // (bufSize) points to the current size of the buffer.
   // (desiredSize) indicates the minimum buffer size required.
   // (copySize) indicates the number of bytes to copy out of the
   // old buffer and into the new one, if a reallocation is necessary
   // Returns true on success, false on failure (out of memory)
   bool EnsureBufferSize(uint8 ** bufPtr, size_t * bufSize, size_t desiredSize, size_t copySize);

   PortableQueue<PortableMessageRef> _outgoingMessages;
   PortableQueue<PortableMessageRef> _incomingMessages;

   PortableDataIO * _io;

   bool _hosed;  // set true on error
   bool _inputEnabled;
   bool _flushOnEmpty;

   PortableMessageRef::ItemPool * _ioMessagePool;
   PortableMessageRef::RefPool * _ioRefPool;
};

#endif
