/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include "iogateway/PortableMessageIOGateway.h"

PortableMessageIOGateway :: PortableMessageIOGateway(PortableDataIO * io, long encoding) :
      AbstractMessageIOGateway(io), _encoding(encoding),
      _sendBuffer(NULL), _sendBufferSize(0), _sendBufferOffset(-1),
      _recvBuffer(NULL), _recvBufferSize(0), _recvBufferOffset(-1)
{
   /* empty */
}

PortableMessageIOGateway :: ~PortableMessageIOGateway() 
{
   delete [] _sendBuffer;
   delete [] _recvBuffer;
}

// Writes some of our outgoing messages to the wire.
// Not guaranteed to write all outgoing messages (it will
// try not to block)
status_t 
PortableMessageIOGateway ::
DoOutput()
{
   if (_hosed) return B_ERROR;

   bool sentBytes = false;
   while(true)
   {
      // First, make sure our outgoing byte-buffer has data.  If it doesn't, fill it with the next outgoing message.
      if (_sendBufferOffset < 0)
      {
         while(true)
         {
            PortableMessageRef nextRef;
            if (_outgoingMessages.RemoveHead(nextRef) == false) 
            {
               if ((_flushOnEmpty)&&(sentBytes)) _io->FlushOutput();
               return B_NO_ERROR;  // nothing more to send, so we're done!
            }
            const PortableMessage * nextSendMsg = nextRef.GetItemPointer();
            _currentSendMessageSize = HeaderSize() + FlattenedMessageBodySize(*nextSendMsg);
            UNLESS(EnsureBufferSize(&_sendBuffer, &_sendBufferSize, _currentSendMessageSize, 0)) {_hosed = true; return B_ERROR;}
            if (FlattenMessage(*nextSendMsg, _sendBuffer, &_sendBuffer[HeaderSize()]) == B_NO_ERROR)
            {
               _sendBufferOffset = 0;
               break;  // go on to sending phase
            }
         }
      }

      // Send whatever data we have pending...
      int attemptSize = _currentSendMessageSize-_sendBufferOffset;
      ssize_t writtenBytes = _io->Write(&_sendBuffer[_sendBufferOffset], attemptSize);
      if (writtenBytes < 0) {_hosed = true; return B_ERROR;}  // trouble!
      _sendBufferOffset += writtenBytes;
      sentBytes = true;
      if (writtenBytes < attemptSize) return B_NO_ERROR;  // assume this means buffers are full; will send the rest later
      if (_sendBufferOffset == _currentSendMessageSize) _sendBufferOffset = -1;  // trigger flattening of next message
   }
}

status_t 
PortableMessageIOGateway ::
DoInput()
{
   status_t errorCode = B_ERROR;

   if (_hosed) return errorCode;

   while(true)
   {
      // First, make sure our incoming byte-buffer has at least enough bytes for the header...
      if (_recvBufferOffset < 0)
      {
         _currentRecvMessageSize = HeaderSize();
         UNLESS(EnsureBufferSize(&_recvBuffer, &_recvBufferSize, _currentRecvMessageSize, 0)) {_hosed = true; return errorCode;}
         _recvBufferOffset = 0;
      }

      // Recv whatever data we have pending...
      int attemptSize = _currentRecvMessageSize-_recvBufferOffset;
      ssize_t readBytes = _io->Read(&_recvBuffer[_recvBufferOffset], attemptSize);
      if (readBytes < 0) {_hosed = true; return errorCode;}  // trouble, connection closed!
      _recvBufferOffset += readBytes;
      if (readBytes < attemptSize) return B_NO_ERROR;  // assume this means input buffers are empty; will recv the rest later
 
      size_t hs = HeaderSize();
      if ((_currentRecvMessageSize == hs)&&(_recvBufferOffset == _currentRecvMessageSize))
      {
         // Now that we have the entire header block, we can determine the size of the message body
         size_t bodySize;
         if (GetBodySize(_recvBuffer, bodySize) != B_NO_ERROR) 
         {  
            // header format error--see if we can recover
            size_t firstValidByte = hs;
            if (RecoverFromInvalidHeader(_recvBuffer, hs) == B_NO_ERROR)
            {
               int j = 0;
               for (size_t i=firstValidByte; i<hs; i++) _recvBuffer[j++] = _recvBuffer[i];
               _recvBufferOffset = j;
               return B_NO_ERROR;
            }
            else
            {
               _hosed = true; 
               return errorCode;
            }
         }
         _currentRecvMessageSize = HeaderSize() + bodySize;
         UNLESS(EnsureBufferSize(&_recvBuffer, &_recvBufferSize, _currentRecvMessageSize, HeaderSize())) {_hosed = true; return errorCode;}
      }
      if (_recvBufferOffset == _currentRecvMessageSize) 
      {
         // Finished receiving message bytes... now reconstruct that bad boy!
         PortableMessage * newMsg = _ioMessagePool ? _ioMessagePool->GetObject() : new (nothrow) PortableMessage;
         if (newMsg == NULL) {WARN_OUT_OF_MEMORY; _hosed = true; return errorCode;}
         status_t ret = UnflattenMessage(*newMsg, _recvBuffer, &_recvBuffer[HeaderSize()], _currentRecvMessageSize-HeaderSize());
         if (ret != B_NO_ERROR)
         {
            if (_ioMessagePool) _ioMessagePool->ReleaseObject(newMsg); 
                           else delete newMsg;  
            if (RecoverFromInvalidMessage(_recvBuffer, _currentRecvMessageSize) == B_NO_ERROR)
            {
               _recvBufferOffset = -1;
               return B_NO_ERROR;
            }
            else
            {
               _hosed = true;
               return errorCode;
            }
         }
         _incomingMessages.AddTail(PortableMessageRef(newMsg, _ioMessagePool, _ioRefPool));
         errorCode = B_NO_ERROR;  // force successful result so that user knows to get his pending messages.
         _recvBufferOffset = -1;  // go back to virgin state for next message
      }
   }
}

// Pass-throughs to the PortableMessage PortableFlattenable
// interface.  Broken out into methods here so that subclasses
// can modify the standard flattening behavior...
size_t 
PortableMessageIOGateway ::
FlattenedMessageBodySize(const PortableMessage & msg) const 
{ 
   return msg.FlattenedSize();
}

// Flattens (msg) into (buffer) and (headerBuf).
// (buffer) is FlattenedMessageSize() bytes long.  This should have header info written to it (including body length!)
// (headerBuf) is HeaderSize() bytes long.  This should have the body data written into it
status_t 
PortableMessageIOGateway ::
FlattenMessage(const PortableMessage & msg, uint8 * headerBuf, uint8 * buffer) const
{
   uint32 * lhb = (uint32 *) headerBuf;
   lhb[0] = B_HOST_TO_LENDIAN_INT32(msg.FlattenedSize());
   lhb[1] = B_HOST_TO_LENDIAN_INT32(_encoding); 
   msg.Flatten(buffer);
   return B_NO_ERROR;
}

// Unflattens (msg) from (buffer) and (headerBuf).
// (buffer) is (bufSize) bytes long.
// (headerBuf) is HeaderSize() bytes long.
status_t 
PortableMessageIOGateway ::
UnflattenMessage(PortableMessage & setMsg, const uint8 * headerBuf, const uint8 * buf, size_t bufSize) const
{
   const uint32 * header = (const uint32 *) headerBuf;
   if (header[0] != B_LENDIAN_TO_HOST_INT32(bufSize)) return B_ERROR;
   if (header[1] != B_LENDIAN_TO_HOST_INT32(_encoding)) return B_ERROR;
   return setMsg.Unflatten(buf, bufSize);
}

// Returns the size of the pre-flattened-message header section, in bytes.
// The default format has an 8-byte header (4 bytes for encoding ID, 4 bytes for message length)
size_t 
PortableMessageIOGateway ::
HeaderSize() const
{
   return 2 * sizeof(uint32);  // one long for the encoding ID, and one long for the body length 
}

// Extracts and returns (in setSize) the dynamic buffer size from the given header.
// Note that the returned size should NOT include the header bytes!
// (headerBuf) is HeaderSize() bytes long.
// Returns B_NO_ERROR iff successful.
status_t 
PortableMessageIOGateway ::
GetBodySize(const uint8 * headerBuf, size_t & setSize) const 
{
   const uint32 * header = (const uint32 *) headerBuf;
   if (B_LENDIAN_TO_HOST_INT32(header[1]) != _encoding) return B_ERROR;
   setSize = (size_t) B_LENDIAN_TO_HOST_INT32(header[0]);
   return B_NO_ERROR;
}

bool 
PortableMessageIOGateway ::
HasBytesToOutput() const
{
   if (_hosed) return false;
   return ((_sendBufferOffset >= 0)||(_outgoingMessages.CountItems() > 0));
}


status_t
PortableMessageIOGateway ::
RecoverFromInvalidHeader(const uint8 * headerBuf, size_t & setFirstValidByte) const
{
   return B_ERROR;  // default behavior:  just give up and die, don't try to recover
}

status_t
PortableMessageIOGateway ::
RecoverFromInvalidMessage(const uint8 * messageBuf, size_t messageBufSize) const
{
   return B_ERROR;  // default behavior:  just give up and die, don't try to recover
}
