/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef OBJECT_POOL_H
#define OBJECT_POOL_H

#include "support/PortableTypeDefs.h"
#include "queue/PortableQueue.h"

//: Support for keeping a "pool" of frequently-used objects, to avoid
//: having to free and re-allocate them all the time.
template <class Object> class ObjectPool
{
public:
   typedef void (*ObjectCallback)(Object * obj, void * userData);

   // (maxPoolSize) is the maximum number of "standby" objects to keep around
   // for recycling at any given time.   
   ObjectPool(uint32 maxPoolSize=100, 
              ObjectCallback recycleCallback = NULL, void * recycleData = NULL,
              ObjectCallback initCallback = NULL, void * initData = NULL) : 
      _maxPoolSize(maxPoolSize),
      _recycleObjectFunc(recycleCallback), _recycleObjectUserData(recycleData),
      _initObjectFunc(initCallback), _initObjectUserData(initData)
   {
      /* empty */
   }

   ~ObjectPool()
   {
      for (int i=_pool.CountItems()-1; i>=0; i--) delete _pool[i];
   }

   // Returns a new object for use (or NULL if no memory available).
   // You are responsible for calling ReleaseObject() on this object
   // when you are done with it.  This object can be used similarly
   // to the new operator.
   Object * GetObject()
   {
      Object * ret = NULL;

      Lock();
      (void) _pool.RemoveHead(ret);
      Unlock();

      if (ret == NULL) ret = new (nothrow) Object;
      if (ret)
      {
         if (_initObjectFunc) _initObjectFunc(ret, _initObjectUserData);
      }
      else WARN_OUT_OF_MEMORY;
      return ret;
   }

   // Adds the given object to our "standby" object list to be recycled.
   // If our list is already at its maximum size, then (obj) will be deleted.
   // (obj) may be NULL.  This method can be used similarly to the delete operator.
   void ReleaseObject(Object * obj, uint32 numItems = 1)
   {
      if (obj)
      {
         if (_recycleObjectFunc) _recycleObjectFunc(obj, _recycleObjectUserData);
         Lock();
         if (_pool.CountItems() < _maxPoolSize)
         {
            _pool.AddTail(obj);
         }
         else 
         {
            if (numItems == 1) delete obj;
                          else delete [] obj;
         }
         Unlock();
      }
   }

   // With this call you can specify a callback function that
   // will be called whenever a new Object is created by this pool. 
   // If you need to do any initialization on the new Object, you can
   // do it here.
   // (userData) will be passed through to the callback function.
   void SetInitObjectCallback(ObjectCallback cb, void * userData)
   {
      _initObjectFunc = cb;
      _initObjectUserData = userData;
   }

   // With this call you can specify a callback function that
   // will be called by ReleaseObject() whenever it is putting an
   // Object into its standby list.
   // If you wish to do any cleanup on the object, you can do it here.
   // (userData) will be passed through to the callback function.
   void SetRecycleObjectCallback(ObjectCallback cb, void * userData)
   {
      _recycleObjectFunc = cb;
      _recycleObjectUserData = userData;
   }

protected:
   virtual void Lock() {/* empty:  subclasses may override this to provide thread safety */}
   virtual void Unlock() {/* empty:  subclasses may override this to provide thread safety */}

private:
   ObjectCallback _initObjectFunc;
   void * _initObjectUserData;
   
   ObjectCallback _recycleObjectFunc;
   void * _recycleObjectUserData;
   
   PortableQueue<Object *> _pool;
   uint32 _maxPoolSize;
};

#endif
