/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <typeinfo>
#include <string.h>
#include <sys/socket.h>

#ifndef __BEOS__
#include <arpa/inet.h>
#endif

#include <netdb.h>

#include "reflector/ReflectServer.h"
#include "netutil/NetworkUtilityFunctions.h"

status_t
ReflectServer ::
AddNewSession(AbstractReflectSessionRef ref, int socket)
{
   AbstractReflectSession * newSession = ref.GetItemPointer();
   newSession->_owner      = this;
   newSession->_socket     = socket;

   if (socket >= 0)
   {
      newSession->_gateway = newSession->CreateGateway(socket);
      if (newSession->_gateway == NULL) 
      {
         newSession->_owner = NULL;
         newSession->_socket = -1;
         return B_ERROR;
      }

      struct sockaddr_in saTempAdd;
#ifdef __BEOS__
      int length = sizeof(sockaddr_in);
#else
      size_t length = sizeof(sockaddr_in);
#endif
      if (getpeername(socket, (struct sockaddr *)&saTempAdd, &length) == 0)
      {
         struct hostent * hePeer = gethostbyaddr((caddr_t)&saTempAdd.sin_addr, sizeof(saTempAdd.sin_addr), AF_INET);
         newSession->_hostName = hePeer ? hePeer->h_name : inet_ntoa(saTempAdd.sin_addr);
         newSession->_port     = ntohs(saTempAdd.sin_port);
      }
      else newSession->_hostName = "<unknown>";
   }
   else
   {
      newSession->_hostName = "<local>";
      newSession->_port     = 0;
   }


   if (_sessions.AddTail(ref))
   {
      _idLookup.Put(newSession->GetSessionIDString(), ref);
      if (newSession->AttachedToServer() == B_NO_ERROR)
      {
         printf("Accepted %s [%s] from [%s:%i], there are now %lu sessions.\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.CountItems());
         return B_NO_ERROR;
      }
      else 
      {
         printf("%s [%s] from [%s:%i] aborted on startup... %lu sessions remain.\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.CountItems());
         _sessions.RemoveTail();
         _idLookup.Remove(newSession->GetSessionIDString());
      }
   }
   else newSession->_owner = NULL;

   return B_ERROR;
}

ReflectServer :: ReflectServer() : _idLookup(CStringHashFunc), _pulseActive(false)
{
   // This call lets the PortableMessage class use its default message pools for its
   // internal purposes.  It's safe to allow this here, since all PortableMessage calls
   // will be made from the same thread.  For a multi-threaded system, don't do this...
   PortableMessage::SetInternalDefaultPools(GetMessagePool(), GetMessageRefPool());

   _idLookup.SetCompareFunction(CStringCompareFunc);
}

ReflectServer :: ~ReflectServer()
{
   // empty
}

void
ReflectServer :: Cleanup()
{
   _idLookup.Clear();
   _lameDuckSessions.MakeEmpty();  // don't delete these, they are ref-counted

   // This silly little contortion is necessary to keep the
   // StorageReflectSession destructors from accessing themselves 
   // in the _sessions list while the _sessions list is being cleared!
   // (i.e. we keep an additional reference of each session around in
   // the (temp) array, so that we can clear _sessions without activating
   // all the destructors.  Then when we clear (temp), the destructors are
   // activated, but the _sessions list is already safely cleared)
   PortableQueue<AbstractReflectSessionRef> temp;
   for (int i=_sessions.CountItems()-1; i>=0; i--) temp.AddTail(_sessions[i]);
   _sessions.MakeEmpty();
   temp.MakeEmpty();
}

status_t
ReflectServer ::
ReadyToRun()
{
   return B_NO_ERROR;
}

const char *
ReflectServer ::
GetServerName() const
{
   return "MUSCLE";
}

status_t 
ReflectServer ::
ServerProcessLoop(ReflectSessionFactoryFunc createSessionFunc, short acceptPort)
{
   // Setup the connection-accepting socket
   struct sockaddr_in saSocket;
   const long trueValue = 1;
   memset(&saSocket, 0, sizeof(saSocket));
   saSocket.sin_family      = AF_INET;
   saSocket.sin_addr.s_addr = htonl(INADDR_ANY);
   saSocket.sin_port        = htons(acceptPort);

   int acceptSocket;
   if ((acceptSocket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
   {
      perror("Error creating accept socket!");
      return B_ERROR;
   }
   (void) setsockopt(acceptSocket, SOL_SOCKET, SO_REUSEADDR, &trueValue, sizeof(long));
   if ((bind(acceptSocket, (struct sockaddr *) &saSocket, sizeof(saSocket)))||(listen(acceptSocket, 20)))
   {
      perror("Error binding/listening on accept socket!");
      closesocket(acceptSocket);
      return B_ERROR;
   }

   if (ReadyToRun() != B_NO_ERROR) 
   {
      printf("Server:  ReadyToRun() failed, aborting.\n");
      return B_ERROR;
   }

   printf("%s Server is active and listening on port %i\n",GetServerName(), acceptPort);
   fd_set readSet;
   fd_set writeSet;
   while(true)
   {
      // Remove any sessions that have experienced I/O errors
      // (Done first since this may have occurred even in ReadyToRun())
      AbstractReflectSession * duck;
      while(_lameDuckSessions.RemoveHead(duck))
      { 
         for (int j=_sessions.CountItems()-1; j>=0; j--)
         {
            AbstractReflectSessionRef nextRef = _sessions[j];  // this must stay in scope, to avoid premature deletions and dangling ptrs
            AbstractReflectSession * next = nextRef.GetItemPointer();
            if (next == duck)
            {
               const char * id = next->GetSessionIDString();
               printf("Closed %s [%s] to [%s:%i], %lu sessions remain.\n", next->GetTypeName(), id, next->GetHostName(), next->GetClientPort(), _sessions.CountItems()-1);
               _idLookup.Remove(id);
               _sessions.RemoveItemAt(j);
               break;
            }
         }
      }

      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);
      FD_SET(acceptSocket, &readSet);

      // Find out which of our sessions wants to write data
      int maxSocket = acceptSocket;
      {
         for (int i=_sessions.CountItems()-1; i>=0; i--)
         {
            AbstractReflectSession * session = _sessions[i].GetItemPointer();
            if (session->_gateway)
            {
               int sessionSocket = session->_socket;
               bool in  = session->_gateway->GetInputEnabled(); 
               bool out = session->_gateway->HasBytesToOutput();     
               if (in)  FD_SET(sessionSocket, &readSet);
               if (out) FD_SET(sessionSocket, &writeSet);
               if (((in)||(out))&&(sessionSocket > maxSocket)) maxSocket = sessionSocket;
            }
         }
      }

      // Calculate timeout for Pulse(), if any
      struct timeval waitTime;
      {
         // If we aren't currently expecting a Pulse(), check to see if we should start one.
         if (_pulseActive == false)
         {
            const struct timeval * timeoutPtr = GetPulseRate();
            if (timeoutPtr) 
            {
               _pulseActive = true;     
               _nextPulseAt = *timeoutPtr;

               struct timeval now; gettimeofday(&now, NULL);
               AddTimeVal(_nextPulseAt, now);  // We want it in absolute time, to avoid "drift" as much as possible.
            }
         }
   
         if (_pulseActive)
         {
            // estimate how long to wait until the pulse time arrives
            waitTime = _nextPulseAt;
            struct timeval now; gettimeofday(&now, NULL);
            SubtractTimeVal(waitTime, now);
            struct timeval zero = {0L, 0L};
            if (IsLessThan(waitTime, zero)) waitTime = zero;
         }
      }
 
      int numBitsSet = select(maxSocket+1, &readSet, &writeSet, NULL, _pulseActive ? &waitTime : NULL);
      if (numBitsSet < 0) 
      {
         closesocket(acceptSocket);
         return B_ERROR;
      }

      if (_pulseActive)
      {
         struct timeval now;
         gettimeofday(&now, NULL);
         if (IsLessThan(_nextPulseAt, now))
         {
            Pulse();
            _pulseActive = false;
         }
      }

      if (FD_ISSET(acceptSocket, &readSet)) 
      {
         // Accept new connection and start up a session for it
#ifdef __BEOS__
         int nLen = sizeof(saSocket);
#else
         size_t nLen = sizeof(saSocket);
#endif
         int newSocket = accept(acceptSocket, (struct sockaddr *)&saSocket, &nLen);
         if (newSocket >= 0)
         {
            AbstractReflectSession * newSession = createSessionFunc();
            if (newSession) 
            {
               AbstractReflectSessionRef newSessionRef(newSession, NULL, NULL);
               if (AddNewSession(newSessionRef, newSocket) != B_NO_ERROR)
               {
                  perror("Error adding new session!");
                  closesocket(newSocket);
               }
            }
            else
            {
               perror("Error, couldn't create new reflector session!");
               closesocket(newSocket);
            }
         }
         else perror("Warning, accept() failed!");
      }

      // Do the session I/Os...
      {
         for (int i=_sessions.CountItems()-1; i>=0; i--)
         {
            AbstractReflectSession & session = *(_sessions[i].GetItemPointer());
            if (session._gateway)
            {
               AbstractMessageIOGateway & gateway = *(session._gateway);
               int socket = session._socket;
               if (socket >= 0)
               {
                  bool readError  = (FD_ISSET(socket, &readSet))&&(gateway.DoInput()   != B_NO_ERROR);
                  bool writeError = (FD_ISSET(socket, &writeSet))&&(gateway.DoOutput() != B_NO_ERROR); 
                  if ((readError)||(writeError)) _lameDuckSessions.AddTail(&session);
               }
            }
         }
      }

      // Let the various sessions handle their messages...
      PortableMessageRef nextMessageRef;
      int numSessions = _sessions.CountItems();
      for (int i=0; i<numSessions; i++)
      {
         AbstractReflectSession * fromSession = _sessions[i].GetItemPointer();
         if (fromSession->_gateway)
         {
            while(fromSession->_gateway->GetNextIncomingMessage(nextMessageRef) == B_NO_ERROR)
               fromSession->MessageReceivedFromGateway(nextMessageRef);
         }
      }
   }
}

void 
ReflectServer ::
Pulse()
{
   // empty
}

const struct timeval *
ReflectServer ::
GetPulseRate()
{
   return NULL;  // default:  never time out
}


PortableQueue<AbstractReflectSessionRef> & 
ReflectServer ::
GetSessionsList()
{
   return _sessions;
}
