/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner"

#ifndef PORTABLE_REFLECT_SERVER_H
#define PORTABLE_REFLECT_SERVER_H

#include "reflector/AbstractReflectSession.h"

//: This class represents a reflection server:  It runs on a centrally located machine,
//: and many clients may connect to it simultaneously.  This server can then redirect messages
//: uploaded by any client to other clients in a somewhat efficient manner.
class ReflectServer
{
public: 
   ReflectServer();
   ~ReflectServer();

   //:The main loop for the message reflection server.
   // Returns B_ERROR if there was a problem setting up.  This method will not return until the server stops running (usually due to an error)
   //  (acceptPort) is the port the server will listen on.
   //  (createSessionFunc) should be a function of the form:
   //     AbstractReflectSession * MyFunc();
   //  that returns a new AbstractReflectSession object to be used for communicating with 
   //  the given socket.  The object this function returns determines the semantics of the server.
   //  Most non-abstract ReflectSession classes come with such a function.
   virtual status_t ServerProcessLoop(ReflectSessionFactoryFunc createSessionFunc, short acceptPort = 2960);

   //: Called after the server is set up, but just before accepting any connections.
   // Should return B_NO_ERROR if it's okay to continue, or B_ERROR to abort.
   // Default implementation returns B_NO_ERROR.
   //
   virtual status_t ReadyToRun();

   // Should be called just before the ReflectServer is to be destroyed.
   // this in a good place to put any cleanup code.  Be sure
   // to call Cleanup() of your parent class as well!
   // (We can't just do this in the destructor, as some cleanup
   // relies on the subclass still being functional, which it isn't
   // when our destructor gets called!)
   virtual void Cleanup();

protected:
   // Adds a new session that uses the given socket for I/O.
   // If (socket) is -1, no TCP connection will be used. 
   virtual status_t AddNewSession(AbstractReflectSessionRef ref, int socket);

   // Returns a human-readable string that describes the type of server that
   // is running.  Default implementation returns "MUSCLE".
   virtual const char * GetServerName() const;

   // May be overridden to return a pointer for a time interval that determines
   // the time interval between calls to our Pulse() method.  By default, this
   // method returns NULL, meaning that Pulse() is never called.  GetPulseRate()
   // will be called repeatedly, so that you can vary the pulse rate if you wish.
   virtual const struct timeval * GetPulseRate();

   // If GetPulseRate() was overridden, then this callback may be called at
   // intervals.  Default implementation does nothing.
   virtual void Pulse();

   // Returns our contained list of active sessions.
   PortableQueue<AbstractReflectSessionRef> & GetSessionsList();

private:
   friend class AbstractReflectSession;

   PortableMessage _centralState;
   PortableQueue<AbstractReflectSessionRef> _sessions;
   PortableHashtable<const char *, AbstractReflectSessionRef> _idLookup;
   PortableQueue<AbstractReflectSession *> _lameDuckSessions;

   struct timeval _nextPulseAt;
   bool _pulseActive;
};

#endif
