/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include "reflector/StorageReflectSession.h"
#include "reflector/StorageReflectUtils.h"

#define DEFAULT_PATH_PREFIX "*/*"  // when we get a path name without a leading '/', prepend this

static PortableMessageRef _blankMessageRef(new PortableMessage, NULL, NULL);

void ResetNodeFunc(StorageReflectSession::DataNode * node, void * userData);
void ResetNodeFunc(StorageReflectSession::DataNode * node, void * userData) {node->Reset();}

StorageReflectSession::DataNodeRef::ItemPool StorageReflectSession::_nodePool(100, ResetNodeFunc);
StorageReflectSession::DataNodeRef::RefPool StorageReflectSession::_nodeRefPool;

StorageReflectSession::StringMatcherRef::ItemPool StorageReflectSession::_stringMatcherPool;
StorageReflectSession::StringMatcherRef::RefPool StorageReflectSession::_stringMatcherRefPool;

void ResetMatcherQueueFunc(PortableQueue<StorageReflectSession::StringMatcherRef> * q, void * userData);
void ResetMatcherQueueFunc(PortableQueue<StorageReflectSession::StringMatcherRef> * q, void * userData) {q->MakeEmpty();}

StorageReflectSession::StringMatcherQueueRef::ItemPool StorageReflectSession::_stringMatcherQueuePool(100, ResetMatcherQueueFunc);
StorageReflectSession::StringMatcherQueueRef::RefPool StorageReflectSession::_stringMatcherQueueRefPool;

AbstractReflectSession * StorageSessionCreationFunc()
{
   AbstractReflectSession * ret = new (nothrow) StorageReflectSession;
   if (ret == NULL) WARN_OUT_OF_MEMORY;
   return ret;   
}

StorageReflectSession ::
StorageReflectSession() : _nextSubscriptionMessage(NULL), _parameters(PR_RESULT_PARAMETERS), _subscriptionsEnabled(true)
{
   // empty
}

StorageReflectSession :: 
~StorageReflectSession() 
{
   Cleanup();
}

status_t
StorageReflectSession ::
AttachedToServer()
{
   // Get the root data node... or add it if it isn't there
   const char * hostname = GetHostName();
   const char * sessionid = GetSessionIDString();

   PortableMessage & state = GetCentralState();
   if (state.FindPointer("root", (void **) &_globalRoot) != B_NO_ERROR)
   {
      // oops, there's no global root!  Gotta add it then
      _globalRoot = GetNewDataNode("", _blankMessageRef);
      if (_globalRoot == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
      if (state.AddPointer("root", _globalRoot) != B_NO_ERROR)
      {
         ReleaseDataNode(_globalRoot);
         return B_ERROR;
      }
   }

   // Is there already a node for our hostname?
   DataNodeRef hostDir;
   if (_globalRoot->GetChild(hostname, hostDir) != B_NO_ERROR)
   {
      // nope.... we'll add one then
      DataNode * hostNode = GetNewDataNode(hostname, _blankMessageRef);
      if (hostNode)
      {
         hostDir = DataNodeRef(hostNode, &_nodePool, &_nodeRefPool);
         _globalRoot->PutChild(hostDir, this, this);
      }
      else {WARN_OUT_OF_MEMORY; Cleanup(); return B_ERROR;}
   }

   // Create a new node for our session (we assume no such
   // node already exists, as session id's are supposed to be unique)
   if (hostDir.GetItemPointer()->HasChild(sessionid))
      printf("WARNING WARNING WARNING:  Non-unique session id [%s] being overwritten!\n", sessionid);

   DataNode * sessionNode = GetNewDataNode(sessionid, _blankMessageRef);
   if (sessionNode)
   { 
      _sessionDir = DataNodeRef(sessionNode, &_nodePool, &_nodeRefPool);
      hostDir.GetItemPointer()->PutChild(_sessionDir, this, this);

      // do subscription notifications here
      PushSubscriptionMessages();
      return B_NO_ERROR;
   }
   else {WARN_OUT_OF_MEMORY; Cleanup(); return B_ERROR;}
}

// Ensures that we don't leave any droppings behind...
void
StorageReflectSession ::
Cleanup()
{
   if (_globalRoot)
   {
      DataNodeRef hostNodeRef;
      if (_globalRoot->GetChild(GetHostName(), hostNodeRef) == B_NO_ERROR)
      {
         DataNode * hostNode = hostNodeRef.GetItemPointer();

         // make sure our session node is gone
         hostNode->RemoveChild(GetSessionIDString(), this, true);

         // If our host node is now empty, it goes too
         if (hostNode->CountChildren() == 0) _globalRoot->RemoveChild(hostNode->GetNodeName(), this, true);
      
         PushSubscriptionMessages();
      }

      // If the global root is now empty, it goes too
      if (_globalRoot->CountChildren() == 0)
      {
         GetCentralState().RemoveName("root");
         ReleaseDataNode(_globalRoot);
      }
      else 
      {
         // Remove all of our subscription-marks from neighbor's nodes
         _subscriptions.DoTraversal((PathMatchCallback)DoSubscribeRefCallbackFunc, this, *_globalRoot, (void *)(-LONG_MAX));
      }
   }
   GetMessagePool()->ReleaseObject(_nextSubscriptionMessage);
}

void 
StorageReflectSession ::
NotifySubscribersThatNodeChanged(DataNode & modifiedNode, bool isBeingRemoved)
{
   PortableHashtableIterator<const char *, uint32> subIter = modifiedNode.GetSubscribers();
   const char * next;
   AbstractReflectSessionRef nRef;
   while(subIter.GetNextKey(next))
   {
      if (GetSessionByID(next, nRef) == B_NO_ERROR) 
      {
         StorageReflectSession * next = (StorageReflectSession *)(nRef.GetItemPointer());
         if ((next != this)||(GetReflectToSelf())) next->NodeChanged(modifiedNode, isBeingRemoved);
      }
   }
}

void
StorageReflectSession ::
NotifySubscribersOfNewNode(DataNode & newNode)
{
   const PortableQueue<AbstractReflectSessionRef> & sessions = GetSessionsList();
   for (int i=sessions.CountItems()-1; i>=0; i--)
   {
      StorageReflectSession * next = (StorageReflectSession *) (sessions[i].GetItemPointer());
      next->NodeCreated(newNode);  // always notify; !Self filtering will be done elsewhere
   }
}

void
StorageReflectSession ::
NodeChanged(DataNode & modifiedNode, bool isBeingRemoved)
{
   if (GetSubscriptionsEnabled())
   {
      if (_nextSubscriptionMessage == NULL) _nextSubscriptionMessage = GetMessageFromPool(PR_RESULT_DATAITEMS);
      if (_nextSubscriptionMessage)
      {
         const char * np = modifiedNode.GetNodePath();
         if (isBeingRemoved) _nextSubscriptionMessage->AddString(PR_NAME_REMOVED_DATAITEMS, np);
                        else _nextSubscriptionMessage->AddMessage(np, modifiedNode.GetData());
      }
      else WARN_OUT_OF_MEMORY;
   }
}

void
StorageReflectSession ::
NodeCreated(DataNode & newNode)
{
   newNode.IncrementSubscriptionRefCount(GetSessionIDString(), _subscriptions.GetMatchCount(newNode));
}

status_t
StorageReflectSession ::
SetDataNode(const PortableString & nodePath, PortableMessageRef dataMsgRef, bool overwrite, bool create)
{
   DataNode * node = _sessionDir.GetItemPointer();
   
   if ((nodePath.Length() > 0)&&(nodePath[0] != '/'))
   {
      int lastSlashPos = -1;
      int slashPos = 0;
      DataNodeRef childNodeRef;
      PortableString nextClause;
      while(slashPos >= 0)
      {
         slashPos = nodePath.IndexOf('/', lastSlashPos+1);
         nextClause = nodePath.Substring(lastSlashPos+1, (slashPos >= 0) ? slashPos : nodePath.Length());
         const char * clause = nextClause.Cstr();
         DataNode * allocedNode = NULL;  // default:  we haven't alloced anybody
         if (node->GetChild(clause, childNodeRef) != B_NO_ERROR)
         {
            if (create)
            {
               // no such child node exists!  Allocate and add it
               allocedNode = GetNewDataNode(clause, _blankMessageRef);
               if (allocedNode)
               {
                  childNodeRef = DataNodeRef(allocedNode, &_nodePool, &_nodeRefPool);
                  node->PutChild(childNodeRef, this, (slashPos < 0) ? NULL : this);  // for last node, notify via SetData(), below
               }
               else {WARN_OUT_OF_MEMORY; continue;}
            }
            else return B_ERROR;
         }

         node = childNodeRef.GetItemPointer();
         if (slashPos < 0)
         {
            if ((overwrite == false)&&(node != allocedNode)) return B_ERROR;
            node->SetData(dataMsgRef, this);
         }
         lastSlashPos = slashPos;
      }
   }
   return B_NO_ERROR;
}

StorageReflectSession :: DataNode *
StorageReflectSession ::
GetDataNode(const PortableString & nodePath) 
{
   DataNode * node = _sessionDir.GetItemPointer();
   if (nodePath.Length() > 0)
   {
      int lastSlashPos = -1;
      int slashPos = 0;
      DataNodeRef childNodeRef;
      PortableString nextClause;

      if (nodePath[0] == '/') 
      {
         node = _globalRoot;
         lastSlashPos=0;
      }
      while(slashPos >= 0)
      {
         slashPos = nodePath.IndexOf('/', lastSlashPos+1);
         nextClause = nodePath.Substring(lastSlashPos+1, (slashPos >= 0) ? slashPos : nodePath.Length());
         const char * clause = nextClause.Cstr();
         if (node->GetChild(clause, childNodeRef) != B_NO_ERROR) return NULL;  // lookup failure (404)
         node = childNodeRef.GetItemPointer();
         if (slashPos < 0) return node;
         lastSlashPos = slashPos;
      }
   }
   return node;
}

void 
StorageReflectSession :: 
MessageReceivedFromGateway(PortableMessageRef msgRef)
{
   const PortableMessage & msg = *msgRef.GetItemPointer();
   switch(msg.what)
   { 
      case PR_COMMAND_SETPARAMETERS:
      {
         bool subscribeQuietly = msg.HasName(PR_NAME_SUBSCRIBE_QUIETLY);
         PortableMessageFieldNameIterator it = msg.GetFieldNameIterator();
         PortableString temp;
         PortableMessage getMsg(PR_COMMAND_GETDATA);
         while(it.GetNextFieldName(temp))
         {
            const char * fn = temp.Cstr();
            bool copyField = true;
            if (strncmp(fn, "SUBSCRIBE:", 10) == 0)
            {
               const char * path = &temp[10];
               PortableString fixPath(path);
               _subscriptions.AdjustStringPrefix(fixPath, DEFAULT_PATH_PREFIX);
               if (_subscriptions.ContainsPathString(fixPath) == false)
               {
                  // This marks any currently existing matching nodes so they know to notify us
                  // It must be done once per subscription path, as it uses per-sub ref-counting
                  WildPathMatcher temp;
                  temp.AddPathString(fixPath);
                  temp.DoTraversal((PathMatchCallback)DoSubscribeRefCallbackFunc, this, *_globalRoot, (void *)1L);

                  _subscriptions.AddPathsFromMatcher(temp);
                  if (subscribeQuietly == false) getMsg.AddString(PR_NAME_KEYS, path);
               }
            }
            else if (strcmp(fn, PR_NAME_REFLECT_TO_SELF) == 0)
            {
               SetReflectToSelf(true);
            }
            else if (strcmp(fn, PR_NAME_DISABLE_SUBSCRIPTIONS) == 0)
            {
               SetSubscriptionsEnabled(false);
            }
            else if (strcmp(fn, PR_NAME_KEYS) == 0)
            {
               _defaultMessageRoute.Clear();
               _defaultMessageRoute.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);
            }
            else if (strcmp(fn, PR_NAME_SUBSCRIBE_QUIETLY) == 0)
            {
               // don't add this to the parameter set; it's just an "argument" for the SUBSCRIBE: fields.
               copyField = false;
            }

            if (copyField) msg.CopyName(fn, _parameters);
         }
         if (getMsg.HasName(PR_NAME_KEYS)) DoGetData(getMsg);  // return any data that matches the subscription
      }
      break;

      case PR_COMMAND_GETPARAMETERS:
      {
         PortableMessage * resultMessage = GetMessageFromPool(_parameters);
         // Add hard-coded params 
         resultMessage->RemoveName(PR_NAME_SESSION_ROOT);
         resultMessage->AddString(PR_NAME_SESSION_ROOT, _sessionDir.GetItemPointer()->GetNodePath());
         if (resultMessage) MessageReceivedFromNeighbor(*this, PortableMessageRef(resultMessage, GetMessagePool(), GetMessageRefPool()), NULL);
                       else WARN_OUT_OF_MEMORY;
      }
      break;

      case PR_COMMAND_REMOVEPARAMETERS:
      {
         StringMatcher matcher;
         const char * nextName;
         PortableString temp;
         PortableQueue<PortableString> toRemove;

         // Compile a list of parameters whose names match any key in the PR_NAME_KEYS field
         for (int i=0; msg.FindString(PR_NAME_KEYS, i, &nextName) == B_NO_ERROR; i++) 
         {
            // Search the parameters message for all parameters that match (nextName)...
            if (matcher.SetPattern(nextName))
            {
               PortableMessageFieldNameIterator it = _parameters.GetFieldNameIterator();
               while(it.GetNextFieldName(temp))
               {
                  if (matcher.Match(temp.Cstr())) toRemove.AddTail(temp);
               }
            }
            else printf("PR_COMMAND_REMOVEPARAMETERS:  Error, StringMatcher couldn't match on [%s]\n",nextName);
         }

         // Now remove them from our parameters list (and do any side effects that go with that)
         for (int j=toRemove.CountItems()-1; j>=0; j--) 
         {
            const char * paramName = toRemove[j].Cstr();
            if (_parameters.RemoveName(paramName) == B_NO_ERROR)
            {
               if (strncmp(paramName, "SUBSCRIBE:", 10) == 0) 
               {
                  PortableString str(&paramName[10]);
                  _subscriptions.AdjustStringPrefix(str, DEFAULT_PATH_PREFIX);
                  if (_subscriptions.RemovePathString(str) == B_NO_ERROR)
                  {
                     // Remove the references from this subscription from all nodes
                     WildPathMatcher temp;
                     temp.AddPathString(str);
                     temp.DoTraversal((PathMatchCallback)DoSubscribeRefCallbackFunc, this, *_globalRoot, (void *)-1L);
                  }
               }
               else if (strcmp(paramName, PR_NAME_REFLECT_TO_SELF) == 0)
               {
                  SetReflectToSelf(false);
               }
               else if (strcmp(paramName, PR_NAME_DISABLE_SUBSCRIPTIONS) == 0)
               {
                  SetSubscriptionsEnabled(true);
               }
            }
         }
      }
      break;

      case PR_COMMAND_SETDATA:
      {
         PortableString nextName;
         PortableMessageFieldNameIterator it = msg.GetFieldNameIterator(B_MESSAGE_TYPE);
         while(it.GetNextFieldName(nextName))
         {
            PortableMessageRef dataMsgRef;
            if (msg.FindMessage(nextName.Cstr(), dataMsgRef) == B_NO_ERROR) SetDataNode(nextName, dataMsgRef);
         }
      }
      break;

      case PR_COMMAND_GETDATA:
         DoGetData(msg);
      break;

      case PR_COMMAND_REMOVEDATA:
      {
         PortableQueue<DataNodeRef> removeSet;
         WildPathMatcher matcher;
         matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, NULL);
         matcher.DoTraversal((PathMatchCallback)RemoveDataCallbackFunc, this, *(_sessionDir.GetItemPointer()), &removeSet);
         for (int i=removeSet.CountItems()-1; i>=0; i--)
         {
            DataNode * next = removeSet[i].GetItemPointer();
            DataNode * parent = next->GetParent();
            if (parent) parent->RemoveChild(next->GetNodeName(), this, true);
         }
      }
      break;

      case PR_RESULT_PARAMETERS:
         // fall-thru
      case PR_RESULT_DATAITEMS:
         printf("Warning, client at [%s] sent me a PR_RESULT_* code.  Bad client!\n", GetHostName());
      break;

      default:  // distribute the message to the appropriate neighbors
      {
         if (msg.HasName(PR_NAME_KEYS, B_STRING_TYPE)) 
         {
            WildPathMatcher matcher;
            matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);
            matcher.DoTraversal((PathMatchCallback)PassMessageCallbackFunc, this, *_globalRoot, &msgRef);
         }
         else if (_parameters.HasName(PR_NAME_KEYS, B_STRING_TYPE)) 
         {
            _defaultMessageRoute.DoTraversal((PathMatchCallback)PassMessageCallbackFunc, this, *_globalRoot, &msgRef);
         }
         else DumbReflectSession::MessageReceivedFromGateway(msgRef);
      }
      break;
   }

   PushSubscriptionMessages();
}

void
StorageReflectSession :: 
DoGetData(const PortableMessage & msg)
{
   PortableMessage * replyMsg = GetMessageFromPool(PR_RESULT_DATAITEMS);
   if (replyMsg)
   {
      PortableMessageRef msgRef(replyMsg, GetMessagePool(), GetMessageRefPool());
      WildPathMatcher matcher;
      matcher.AddPathsFromMessage(PR_NAME_KEYS, msg, DEFAULT_PATH_PREFIX);  
      matcher.DoTraversal((PathMatchCallback)GetDataCallbackFunc, this, *_globalRoot, replyMsg);
      MessageReceivedFromNeighbor(*this, msgRef, NULL);
   }
   else WARN_OUT_OF_MEMORY;
}

void
StorageReflectSession ::
PushSubscriptionMessages()
{
   // Send out any subscription results that were generated...
   const PortableQueue<AbstractReflectSessionRef> & sessionList = GetSessionsList();
   for (int i=sessionList.CountItems()-1; i>=0; i--)
   {
      StorageReflectSession * nextSession = (StorageReflectSession *) sessionList[i].GetItemPointer();
      PortableMessage * subMsg = nextSession->_nextSubscriptionMessage;
      while(subMsg) 
      {
         nextSession->_nextSubscriptionMessage = NULL;  /* In case FromNeighbor() wants to add more suscriptions... */
         nextSession->MessageReceivedFromNeighbor(*this, PortableMessageRef(subMsg, GetMessagePool(), GetMessageRefPool()), NULL);
         subMsg = nextSession->_nextSubscriptionMessage;
      }
   }
}

int
StorageReflectSession ::
PassMessageCallback(DataNode & node, void * userData)
{
   DataNode * n = &node;
   while(n->GetDepth() > 2) n = n->GetParent();  // go up to session level...
   AbstractReflectSessionRef sref;
   if (GetSessionByID(n->GetNodeName(), sref) == B_NO_ERROR) 
   {
      sref.GetItemPointer()->MessageReceivedFromNeighbor(*this, *((PortableMessageRef *)userData), &node);
   }
   return 2; // This causes the traversal to immediately skip to the next session
}


int
StorageReflectSession ::
DoSubscribeRefCallback(DataNode & node, void * userData)
{
   node.IncrementSubscriptionRefCount(GetSessionIDString(), (long) userData);
   return node.GetDepth();  // continue traversal as usual
}

int
StorageReflectSession ::
GetDataCallback(DataNode & node, void * userData)
{
   if (GetReflectToSelf() == false)
   {
      // Make sure (node) isn't part of our own tree!
      // If it is, move immediately to the next session
      const DataNode * n = &node;
      while(n->GetDepth() > 2) n = n->GetParent();
      AbstractReflectSessionRef ref;
      if ((GetSessionByID(n->GetNodeName(), ref) == B_NO_ERROR)&&(ref.GetItemPointer() == this)) return 2;  
   }
   ((PortableMessage *)userData)->AddMessage(node.GetNodePath(), node.GetData());
   return node.GetDepth();  // continue traveral as usual
}

int
StorageReflectSession ::
RemoveDataCallback(DataNode & node, void * userData)
{
   if (node.GetDepth() > 2)  // ensure that we never remove host nodes or session nodes this way
   {
      DataNodeRef nodeRef;
      if (node.GetParent()->GetChild(node.GetNodeName(), nodeRef) == B_NO_ERROR) 
      {
         ((PortableQueue<DataNodeRef> *)userData)->AddTail(nodeRef);
         return node.GetDepth()-1;  // no sense in recursing down a node that we're going to delete anyway
      }
   }
   return node.GetDepth();
}


StorageReflectSession :: DataNode ::
DataNode() : _children(CStringHashFunc), _subscribers(CStringHashFunc)
{
   _children.SetCompareFunction(CStringCompareFunc);
   _subscribers.SetCompareFunction(CStringCompareFunc);
}

void 
StorageReflectSession :: DataNode ::
Init(const char * name, PortableMessageRef initData)
{
   _nodeName = name;
   _parent   = NULL;
   _depth    = 0;
   _data     = initData;
}

void
StorageReflectSession :: DataNode ::
Reset()
{
   _children.Clear();
   _subscribers.Clear();
   _parent = NULL;
   _depth = 0;
   _data.Reset();
}

void
StorageReflectSession :: DataNode :: 
IncrementSubscriptionRefCount(const char * sessionID, long delta)
{
   if (delta > 0)
   {
      uint32 res = 0;
      (void)_subscribers.Get(sessionID, res);
      _subscribers.Put(sessionID, res+delta);
   }
   else if (delta < 0)
   {
      uint32 res = 0;
      if (_subscribers.Get(sessionID, res))
      {
         uint32 decBy = (uint32) -delta;
         if (decBy >= res) _subscribers.Remove(sessionID);
                      else _subscribers.Put(sessionID, res - decBy);
      }
   }
}

PortableHashtableIterator<const char *, uint32> 
StorageReflectSession :: DataNode ::
GetSubscribers() const
{
   return _subscribers.GetIterator();;
}

void
StorageReflectSession :: DataNode ::
PutChild(DataNodeRef & node, StorageReflectSession * optNotifyWithOnSetParent, StorageReflectSession * optNotifyChangedData)
{
   DataNode * child = node.GetItemPointer();
   child->SetParent(this, optNotifyWithOnSetParent);
   _children.Put(child->_nodeName.Cstr(), node);
   if (optNotifyChangedData) optNotifyChangedData->NotifySubscribersThatNodeChanged(*child, false);
}

StorageReflectSession :: DataNode *
StorageReflectSession :: DataNode ::
GetParent() const
{
   return _parent;
}

int
StorageReflectSession :: DataNode ::
GetDepth() const
{
   return _depth;
}

void
StorageReflectSession :: DataNode ::
SetParent(DataNode * parent, StorageReflectSession * optNotifyWith)
{
   if ((_parent)&&(parent)) printf("Warning, overwriting previous parent of node [%s]\n", GetNodeName());
   _parent = parent;
   if (_parent == NULL) _subscribers.Clear();
   _depth = 0;

   // Calculate node path and node depth
   if (_parent)
   {
      // Calculate the total length that our node path string will be
      uint32 pathLen = 0;
      {
         const DataNode * node = this;
         while(node->_parent) 
         {
            pathLen += 1 + node->_nodeName.Length();  // the 1 is for the slash
            node = node->_parent;
            _depth++;
         }
      }

      char * aBuf = NULL;
      const int staticAllocSize = 200;
      char sBuf[staticAllocSize];  // try to do this without a dynamic allocation...
      if (pathLen >= staticAllocSize)  // but do a dynamic allocation if we have to (should be rare)
      {
         aBuf = new (nothrow) char[pathLen+1];
         if (aBuf == NULL) 
         { 
            WARN_OUT_OF_MEMORY; 
            _nodePath = "<error>"; 
            return;
         }
      }

      char * writeAt = (aBuf ? aBuf : sBuf) + pathLen;  // points to last char in buffer
      *writeAt = '\0';  // terminate the string first (!)
      const DataNode * node = this;
      while(node->_parent)
      {
         int len = node->_nodeName.Length();
         writeAt -= len;
         memcpy(writeAt, node->_nodeName.Cstr(), len);
         *(--writeAt) = '/';
         node = node->_parent;
      }
      ASSERT(writeAt == (aBuf ? aBuf : sBuf), "GetNodePath has a bug!");
 
      _nodePath = (aBuf ? aBuf : sBuf);
      delete [] aBuf;

      if (optNotifyWith) optNotifyWith->NotifySubscribersOfNewNode(*this);
   }
}

status_t
StorageReflectSession :: DataNode ::   
GetChild(const char * key, DataNodeRef & returnChild) const
{
   return _children.Get(key, returnChild) ? B_NO_ERROR : B_ERROR;
}

const char * 
StorageReflectSession :: DataNode ::
GetNodePath() const
{
   return _nodePath.Cstr();
}


bool
StorageReflectSession :: DataNode ::   
HasChild(const char * key) const
{
   return _children.ContainsKey(key);
}

status_t
StorageReflectSession :: DataNode ::   
RemoveChild(const char * key, StorageReflectSession * optNotifyWith, bool recurse)
{
   DataNodeRef childRef;
   if (_children.Get(key, childRef))
   {
      DataNode * child = childRef.GetItemPointer();
      if (optNotifyWith) optNotifyWith->NotifySubscribersThatNodeChanged(*child, true);
      if (recurse)
      { 
         while(child->CountChildren() > 0)
         {
            DataNodeRefIterator it = child->GetChildIterator();
            const char * name;
            it.GetNextKey(name);
            child->RemoveChild(name, optNotifyWith, recurse);
         }
      }

      child->SetParent(NULL, optNotifyWith);
      _children.Remove(key, childRef);
      return B_NO_ERROR;
   }
   else return B_ERROR;
}

uint32
StorageReflectSession :: DataNode ::   
CountChildren() const
{
   return _children.Size();
}

const char *
StorageReflectSession :: DataNode ::   
GetNodeName() const
{
   return _nodeName.Cstr();
}

void
StorageReflectSession :: DataNode ::   
SetData(PortableMessageRef data, StorageReflectSession * optNotifyWith)
{
   _data = data;
   if (optNotifyWith) optNotifyWith->NotifySubscribersThatNodeChanged(*this, false);
}

PortableMessageRef 
StorageReflectSession :: DataNode ::
GetData()
{
   return _data;
}

StorageReflectSession :: DataNode ::
~DataNode() 
{
//printf("Deleting datanode %p/%s\n",this,_nodePath.Cstr());
   // empty
}

StorageReflectSession :: DataNodeRefIterator
StorageReflectSession :: DataNode ::
GetChildIterator() const
{
   return _children.GetIterator();
}

StorageReflectSession :: WildPathMatcher ::
WildPathMatcher()
{
   // empty
}

StorageReflectSession :: WildPathMatcher ::
~WildPathMatcher()
{
   // empty
}

void 
StorageReflectSession :: WildPathMatcher :: 
Clear()
{
   _parsers.MakeEmpty();
   _pathStrings.MakeEmpty();
}

void
StorageReflectSession :: WildPathMatcher ::
AdjustStringPrefix(PortableString & path, const char * optPrepend) const
{
   if (path.Length() > 0)
   {
           if (path[0] == '/') path = path.Substring(1);
      else if (optPrepend)     path = optPrepend + ("/" + path);
   }
}

status_t 
StorageReflectSession :: WildPathMatcher ::
AddPathString(const PortableString & path)
{
   if (path.Length() > 0) 
   {
      PortableQueue<StringMatcherRef> * newQ = _stringMatcherQueuePool.GetObject();
      if (newQ == NULL) {WARN_OUT_OF_MEMORY; return B_ERROR;}
      StringMatcherQueueRef qRef(newQ, &_stringMatcherQueuePool, &_stringMatcherQueueRefPool);

      PortableString temp;
      int lastSlashPos = -1;
      int slashPos = 0;
      while(slashPos >= 0)
      {
         slashPos = path.IndexOf('/', lastSlashPos+1);
         temp = path.Substring(lastSlashPos+1, (slashPos >= 0) ? slashPos : path.Length());
         StringMatcher * sm = NULL;  // default, same as "*"
         if (strcmp(temp.Cstr(), "*"))
         {
            sm = _stringMatcherPool.GetObject();
            if (sm == NULL) WARN_OUT_OF_MEMORY;
            sm->SetPattern(temp.Cstr());
         }
         newQ->AddTail(StringMatcherRef(sm, &StorageReflectSession::_stringMatcherPool, &StorageReflectSession::_stringMatcherRefPool));
         lastSlashPos = slashPos;
      }
      _parsers.AddTail(qRef);
      _pathStrings.AddTail(path);
      return B_NO_ERROR;
   }
   else return B_ERROR;
}

status_t 
StorageReflectSession :: WildPathMatcher ::
RemovePathString(const PortableString & path)
{
   int which = _pathStrings.IndexOf(path);
   if (which >= 0) 
   {
      _parsers.RemoveItemAt(which);
      _pathStrings.RemoveItemAt(which);
      return B_NO_ERROR;
   }
   else printf("ERROR, COULDN'T REMOVE SUBSCRIBE STRING! [%s]\n", path.Cstr());
   return B_ERROR;
}

bool
StorageReflectSession :: WildPathMatcher ::
ContainsPathString(const PortableString & w) const
{
   return (_pathStrings.IndexOf(w) >= 0);
}

status_t 
StorageReflectSession :: WildPathMatcher ::
AddPathsFromMessage(const char * fieldName, const PortableMessage & msg, const char * prependIfNoLeadingSlash)
{
   status_t ret = B_NO_ERROR;
   PortableString str;
   for (int i=0; msg.FindString(fieldName, i, str) == B_NO_ERROR; i++) 
   {
      AdjustStringPrefix(str, prependIfNoLeadingSlash);
      if (AddPathString(str) != B_NO_ERROR) ret = B_ERROR;
   }
   return ret;
}

void 
StorageReflectSession :: WildPathMatcher ::
AddPathsFromMatcher(const WildPathMatcher & matcher)
{
   for (int i=matcher._parsers.CountItems()-1; i>=0; i--)
   {
      _pathStrings.AddTail(*matcher._pathStrings.GetItemPointer(i));
      _parsers.AddTail(*matcher._parsers.GetItemPointer(i));
   }
}

bool
StorageReflectSession :: WildPathMatcher ::
MatchesNode(DataNode & node) const
{
   for (int i=_parsers.CountItems()-1; i>=0; i--) if (PathMatches(node, i)) return true;
   return false;
}

int
StorageReflectSession :: WildPathMatcher ::
GetMatchCount(DataNode & node) const
{
   int matchCount = 0;
   for (int i=_parsers.CountItems()-1; i>=0; i--) if (PathMatches(node, i)) matchCount++;
   return matchCount;
}

bool
StorageReflectSession :: WildPathMatcher ::
PathMatches(DataNode & node, int whichPath) const
{
   PortableQueue<StringMatcherRef> * nextSubscription = _parsers[whichPath].GetItemPointer();
   int pd = nextSubscription->CountItems();
   if (pd == node.GetDepth())  // only paths with the same number of clauses as the node's path can ever match
   {
      DataNode * travNode = &node;
      bool match = true;  // optimistic default
      for (int j=nextSubscription->CountItems()-1; j>=0; j--,travNode=travNode->GetParent())
      {
         StringMatcher * nextMatcher = nextSubscription->GetItemPointer(j)->GetItemPointer();
         if ((nextMatcher)&&(nextMatcher->Match(travNode->GetNodeName()) == false)) 
         {
            match = false;
            break; 
         }
      }
      if (match) return true;
   }
   return false;
}

void
StorageReflectSession :: WildPathMatcher ::
DoTraversal(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData)
{
   (void) DoTraversalAux(cb, This, node, userData, node.GetDepth());
}

int 
StorageReflectSession :: WildPathMatcher ::
DoTraversalAux(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData, int rootDepth)
{
   DataNodeRef nextChildRef;
   DataNodeRefIterator it = node.GetChildIterator();
   int depth = node.GetDepth();
   while(it.GetNextValue(nextChildRef)) 
   {
      DataNode & nextChild = *nextChildRef.GetItemPointer();
      const char * nextChildName = nextChild.GetNodeName();
      bool matched  = false;  // set if we have called the callback on this child already
      bool recursed = false;  // set if we have recursed to this child already

      // Try all parsers and see if any of them match at this level
      int numParsers = _parsers.CountItems();
      for (int i=0; i<numParsers; i++)
      {
         PortableQueue<StringMatcherRef> * nextQueue = _parsers[i].GetItemPointer();
         int numClausesInParser = nextQueue->CountItems();
         if (numClausesInParser > depth-rootDepth)
         {
            StringMatcher * nextMatcher = nextQueue->GetItemPointer(depth-rootDepth)->GetItemPointer();
            if ((nextMatcher == NULL)||(nextMatcher->Match(nextChildName)))
            {
               // A match!  Now, depending on whether this match is the
               // last clause in the path or not, we either do the callback or descend.
               // But we make sure not to do either of these things more than once per node.

               if (depth == rootDepth+numClausesInParser-1)
               {
                  if (matched == false)
                  {
                     // when there is more than one string being used to match,
                     // it's possible that two or more strings can "conspire"
                     // to match a node even though any given string doesn't match it.
                     // For example, if we have the match-strings:
                     //    /j*/k*
                     //    /k*/j*
                     // The node /jeremy/jenny would match, even though it isn't
                     // specified by any of the subscription strings.  This is bad.
                     // So for multiple match-strings, we do an additional check 
                     // to make sure there is a WildPathMatcher for this node.
                     if ((numParsers == 1)||(MatchesNode(nextChild)))
                     {
                        int nextDepth = cb(This, nextChild, userData);
                        if (nextDepth < nextChild.GetDepth()-1) return nextDepth;
                        matched = true;
                        if (recursed) break;  // done both possibile actions, so be lazy
                     }
                  }
               }
               else
               {
                  if (recursed == false)
                  {
                     // If we match a non-terminal clause in the path, recurse to the child.
                     int nextDepth = DoTraversalAux(cb, This, nextChild, userData, rootDepth);
                     if (nextDepth < nextChild.GetDepth()-1) return nextDepth;
                     recursed = true;
                     if (matched) break;  // done both possible actions, so be lazy
                  }
               }
            }
         }
      }
   }
   return node.GetDepth();
}

StorageReflectSession::DataNode *
StorageReflectSession ::
GetNewDataNode(const char * name, PortableMessageRef initialValue)
{
   DataNode * n = _nodePool.GetObject();
   n->Init(name, initialValue);
   return n;
}

void
StorageReflectSession ::
ReleaseDataNode(DataNode * node)
{
   _nodePool.ReleaseObject(node);
}
