/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */

//! author="Jeremy Friesner" 

/*******************************************************************************
/
/   File:      Rect.h
/
/   Description:    Portable version of Be's BRect class
/
*******************************************************************************/

#ifndef PORTABLE_RECT_H
#define PORTABLE_RECT_H

#include <math.h>
#include "support/PortablePoint.h"

/*-----------------------------------------------------------------------*/
/*----- PortableRect class ----------------------------------------------*/

//:A portable version of Be's BRect class.
class PortableRect : public PortableFlattenable
{
public:
   float left;
   float top;
   float right;
   float bottom;

   PortableRect() {left = top = 0.0f;  right = bottom = -1.0f;}
   PortableRect(float l, float t, float r, float b) {Set(l,t,r,b);}
   PortableRect(PortablePoint leftTop, PortablePoint rightBottom)
   {
      Set(leftTop.x, leftTop.y, rightBottom.x, rightBottom.y);
   }

   void Set(float l, float t, float r, float b)
   {
      left   = l;
      top    = t;
      right  = r;
      bottom = b;
   }

   void PrintToStream() const
   {
      printf("PortableRect: leftTop=(%f,%f) rightBottom=(%f,%f)\n", left, top, right, bottom);
   }

/* PortablePoint selectors */
   PortablePoint LeftTop() const {return PortablePoint(left, top);}
   PortablePoint RightBottom() const {return PortablePoint(right, bottom);}
   PortablePoint LeftBottom() const {return PortablePoint(left, bottom);}
   PortablePoint RightTop() const {return PortablePoint(right, top);}

/* PortablePoint setters */
   void SetLeftTop(const PortablePoint p) {left = p.x; top = p.y;}
   void SetRightBottom(const PortablePoint p) {right = p.x; bottom = p.y;}
   void SetLeftBottom(const PortablePoint p) {left = p.x; bottom = p.y;}
   void SetRightTop(const PortablePoint p) {right = p.x; top = p.y;}

/* transformation */
   void InsetBy(PortablePoint p) {InsetBy(p.x, p.y);}
   void InsetBy(float dx, float dy) {left += dx; top += dy; right -= dx; bottom -= dy;}
   void OffsetBy(PortablePoint p) {OffsetBy(p.x, p.y);}
   void OffsetBy(float dx, float dy) {left += dx; top += dy; right += dx; bottom += dy;}
   void OffsetTo(PortablePoint p) {OffsetTo(p.x, p.y);}
   void OffsetTo(float x, float y) {right = x + Width(); bottom = y + Height(); left = x; top = y;}

/* comparison */
   bool operator==(PortableRect r) const {return (left == r.left)&&(top == r.top)&&(right == r.right)&&(bottom == r.bottom);}
   bool operator!=(PortableRect r) const {return !(r == *this);}

/* intersection and union */
   PortableRect operator&(PortableRect r) const 
   {
      PortableRect ret(*this);
      if (ret.left   < r.left)   ret.left   = r.left;
      if (ret.right  > r.right)  ret.right  = r.right;
      if (ret.top    < r.top)    ret.top    = r.top;
      if (ret.bottom > r.bottom) ret.bottom = r.bottom;
      return ret;
   }

   PortableRect operator|(PortableRect r) const 
   {
      PortableRect ret(*this);
      if (r.left   < ret.left)   ret.left   = r.left;
      if (r.right  > ret.right)  ret.right  = r.right;
      if (r.top    < ret.top)    ret.top    = r.top;
      if (r.bottom > ret.bottom) ret.bottom = r.bottom;
      return ret;
   }

/* utilities */
   bool Intersects(PortableRect r) const {return (r&(*this)).IsValid();}

   bool IsValid() const {return ((Width() >= 0.0f)&&(Height() >= 0.0f));}
   float Width() const {return right - left;}
   int32 IntegerWidth() const {return (int32)ceil(Width());}
   float Height() const {return bottom - top;}
   int32 IntegerHeight() const {return (int32)ceil(Height());}
   bool Contains(PortablePoint p) const
   {
      return ((p.x >= left)&&(p.x <= right)&&(p.y >= top)&&(p.y <= bottom));
   }

   bool Contains(PortableRect p) const
   {
      return ((Contains(p.LeftTop()))&&(Contains(p.RightTop()))&&
              (Contains(p.LeftBottom()))&&(Contains(p.RightBottom())));
   }

   virtual bool IsFixedSize() const {return true;}
   virtual type_code TypeCode() const {return B_RECT_TYPE;}
   virtual size_t FlattenedSize() const {return 4 * sizeof(float);}

   virtual void Flatten(uint8 * buffer) const
   {
      float * buf = (float *) buffer;
      buf[0] = B_HOST_TO_LENDIAN_FLOAT(left);
      buf[1] = B_HOST_TO_LENDIAN_FLOAT(top);
      buf[2] = B_HOST_TO_LENDIAN_FLOAT(right);
      buf[3] = B_HOST_TO_LENDIAN_FLOAT(bottom);
   }

   virtual status_t Unflatten(const uint8 * buffer, size_t size)
   {
      if (size >= FlattenedSize())
      {
         float * buf = (float *) buffer;
         left   = B_LENDIAN_TO_HOST_FLOAT(buf[0]);
         top    = B_LENDIAN_TO_HOST_FLOAT(buf[1]);
         right  = B_LENDIAN_TO_HOST_FLOAT(buf[2]);
         bottom = B_LENDIAN_TO_HOST_FLOAT(buf[3]);
         return B_NO_ERROR;
      }
      else return B_ERROR;
   }
};

#endif 
