/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>

#include <app/Looper.h>
#include "dataio/TCPSocketDataIO.h"
#include "iogateway/PortableMessageIOGateway.h"
#include "reflector/StorageReflectConstants.h"
#include "besupport/MessageTransceiverThread.h"
#include "netutil/NetworkUtilityFunctions.h"

#define TEST(x) if ((x) != B_NO_ERROR) printf("Test failed, line %i\n",__LINE__)

#define METHOD_MANUAL      0
#define METHOD_AUTOMATIC   1
#define METHOD_ACCEPT_ONE  2
#define METHOD_ACCEPT_MANY 3

// This BeOS-specific program is used to test the muscled server.  It is more-or-less
// functionally equivalent to the portablereflectclient, but exercises the BeOS-specific
// functionality found in the besupport class as well. 
class MyLooper : public BLooper
{
public:
   MyLooper(int method) : _method(method) {}

   virtual void MessageReceived(BMessage * msg)
   {
      switch(msg->what)
      {
         case PORTABLE_MESSAGES_RECEIVED:
         {
            PortableMessageRef ref; 
            while(_transThread->GetNextIncomingMessage(ref) == B_NO_ERROR)
            {
               printf("Heard message from server:-----------------------------------\n");
               ref.GetItemPointer()->PrintToStream();
               printf("-------------------------------------------------------------\n");
            }
         }
         break;

         case '_BYE':
            printf("MyLooper: Connection to server was closed!\n");
            fflush(stdout);
            if (_method != METHOD_ACCEPT_MANY) exit(0);
         break;

         case 'HELO':
            printf("MyLooper: Connection to server is ready!\n");
            fflush(stdout);
         break;

         default:
            printf("MyLooper:  Received unknown BMessage:\n");
            msg->PrintToStream();
         break;
      }
   }

   MessageTransceiverThread * _transThread;
   int _method;
};

MessageTransceiverThread * CreateTransceiverThread(const char * hostName, int port, MyLooper * looper, int method)
{
   BMessenger messenger(looper);
   BMessage connectMsg('HELO');
   BMessage disconnectMsg('_BYE');
   MessageTransceiverThread * mtt = NULL;

   looper->Run();
   switch(method)
   {
      case METHOD_MANUAL:
      {
         int s = Connect(hostName, port, "trc");
         if (s >= 0)
         {
            PortableMessageIOGateway * gw = new PortableMessageIOGateway(new TCPSocketDataIO(s, false));
            mtt = new MessageTransceiverThread(messenger);
            if (mtt->StartThread(gw, s, &disconnectMsg) != B_NO_ERROR)
            {
               printf("Error starting thread!\n");
               delete mtt;
               delete gw;
               mtt = NULL;
            }
         }
      }
      break;

      case METHOD_AUTOMATIC:
         mtt = new MessageTransceiverThread(messenger);
         if (mtt->StartConnectThread(hostName, port, &connectMsg, &disconnectMsg) != B_NO_ERROR)
         {
            printf("Error starting auto-connect thread!\n");
            delete mtt;
            mtt = NULL;
         }
         else printf("Connect asynchronously to [%s:%i]\n", hostName, port);
      break;

      case METHOD_ACCEPT_ONE:
         mtt = new MessageTransceiverThread(messenger);
         if (mtt->StartAcceptThread(port, false, &connectMsg, &disconnectMsg) != B_NO_ERROR)
         {
            printf("Error starting auto-accept-one thread!\n");
            delete mtt;
            mtt = NULL;
         }
         printf("Accepting one connection on port %i\n", mtt->GetPort());
      break;

      case METHOD_ACCEPT_MANY:
         mtt = new MessageTransceiverThread(messenger);
         if (mtt->StartAcceptThread(port, true, &connectMsg, &disconnectMsg) != B_NO_ERROR)
         {
            printf("Error starting auto-accept-many thread!\n");
            delete mtt;
            mtt = NULL;
         }
         printf("Accepting more than one connection on port %i\n", mtt->GetPort());
      break;

      default:
         printf("CreateTransceiverThread(): unknown method %i\n",method);
      break;
   }
   if (mtt) looper->_transThread = mtt;
   return mtt;
}

void main(int argc, char ** argv)
{
   char * hostName = "localhost";
   int port = 0;

   int command = -1;  
   if (argc > 1) 
   {
      if (strcmp(argv[1], "-connect") == 0)
      {
         command = METHOD_AUTOMATIC;
         if (argc > 2) hostName = argv[2];
         if (argc > 3) port = atoi(argv[3]);
         if (port == 0) port = 2960;
      } 
      else if (strcmp(argv[1], "-connectsync") == 0)
      {
         command = METHOD_MANUAL;
         if (argc > 2) hostName = argv[2];
         if (argc > 3) port = atoi(argv[3]);
         if (port == 0) port = 2960;
      }
      else if (strcmp(argv[1], "-acceptone") == 0)
      {
         command = METHOD_ACCEPT_ONE;
         if (argc > 2) port = atoi(argv[2]);
      }
      else if (strcmp(argv[1], "-acceptmany") == 0)
      {
         command = METHOD_ACCEPT_MANY;
         if (argc > 2) port = atoi(argv[2]);
      }
   }

   if (command != -1)
   {
      MyLooper * looper = new MyLooper(command);
      MessageTransceiverThread * mtt = CreateTransceiverThread(hostName, port, looper, command);
      if (mtt)
      {
         char text[1000];
         bool keepGoing = true;
         while(keepGoing)
         {
            gets(text);
            printf("You typed: [%s]\n",text);
            int len = strlen(text);
            if (len > 0)
            {
               bool send = true;
               PortableMessage * msg = new PortableMessage;
               PortableMessageRef ref(msg, NULL, NULL);

               switch(text[0])
               {
                  case 'm':
                     msg->what = 'umsg';
                     if (text[1]) msg->AddString(PR_NAME_KEYS, &text[2]);
                     msg->AddString("info", text[1] ? "This is a directed user message" : "This is a default-directed user message");
                  break;

                  case 'M':
                     msg->what = PR_COMMAND_SETPARAMETERS;
                     msg->AddString(PR_NAME_KEYS, &text[2]);
                  break;

                  case 's':
                     msg->what = PR_COMMAND_SETDATA;
                     msg->AddMessage(&text[2], PortableMessage('HELO')); 
                  break;

                  case 'g':
                     msg->what = PR_COMMAND_GETDATA;
                     msg->AddString(PR_NAME_KEYS, &text[2]);
                  break;

                  case 'G':
                     msg->what = PR_COMMAND_GETDATA;
                     msg->AddString(PR_NAME_KEYS, "j*/k*");
                     msg->AddString(PR_NAME_KEYS, "k*/j*");
                  break;

                  case 'q':
                     keepGoing = send = false;
                  break;

                  case 'p':
                     msg->what = PR_COMMAND_SETPARAMETERS;
                     msg->AddString(&text[2], "");
                  break;

                  case 'P':
                     msg->what = PR_COMMAND_GETPARAMETERS;
                  break;

                  case 'd':
                     msg->what = PR_COMMAND_REMOVEDATA;
                     msg->AddString(PR_NAME_KEYS, &text[2]);
                  break;

                  case 'D':
                     msg->what = PR_COMMAND_REMOVEPARAMETERS;
                     msg->AddString(PR_NAME_KEYS, &text[2]);
                  break;

                  case 't':
                  {
                     // test all data types
                     msg->what = 1234;
                     msg->AddString("String", "this is a string");
                     msg->AddInt8("Int8", 123);
                     msg->AddInt8("-Int8", -123);
                     msg->AddInt16("Int16", 1234);
                     msg->AddInt16("-Int16", -1234);
                     msg->AddInt32("Int32", 12345);
                     msg->AddInt32("-Int32", -12345);
                     msg->AddInt64("Int64", 123456789);
                     msg->AddInt64("-Int64", -123456789);
                     msg->AddBool("Bool", true);
                     msg->AddBool("-Bool", false);
                     msg->AddFloat("Float", 1234.56789f);
                     msg->AddFloat("-Float", -1234.56789f);
                     msg->AddDouble("Double", 1234.56789);
                     msg->AddDouble("-Double", -1234.56789);
                     msg->AddPointer("Pointer", msg);
                     msg->AddFlat("Flat", *msg);
                     char data[] = "This is some data";
                     msg->AddData("Flat", B_RAW_TYPE, data, sizeof(data));
                  }
                  break;

                  default:
                     printf("Sorry, wot?\n");
                     send = false;
                  break;
               }

               if (send) 
               {
                  printf("Sending message...\n");
//                  ref.GetItemPointer()->PrintToStream();
                  mtt->SendOutgoingMessage(ref);
               }
            }
         }

         printf("Waiting for all pending messages to be sent...\n");
         mtt->WaitForAllMessagesToBeSent();
         printf("Shutting down looper...\n");
         if (looper->Lock()) looper->Quit();
         delete mtt;
      }
      else printf("Couldn't set up session!\n");
   }
   else
   {
      printf("Usage:  testreflectclient -connect [hostname=localhost] [port=2960]\n");
      printf("        testreflectclient -connectsync [hostname=localhost] [port=2960]\n");
      printf("        testreflectclient -acceptone [port=0]\n");
      printf("        testreflectclient -acceptmany [port=0]\n");
   }
   printf("Test client exiting, bye!\n");
}

